@@ -1,11 +1,7 @@
-// external imports
-use ndarray::Array;
-use ndarray::Array2;
-// crate imports
 use crate::network::{
     net::unroll, EdgeLike, Fabricator, NetLike, NodeLike, Recurrent, StatefulFabricator,
 };
-// std imports
+use nalgebra::{DMatrix, DVector};
 use std::collections::HashMap;
 
 pub struct FeedForwardMatrixFabricator;
@@ -27,7 +23,7 @@ where
         assert!(unrolled.inputs().len() - net.inputs().len() == memory);
 
         Ok(super::evaluator::RecurrentMatrixEvaluator {
-            internal: Array::zeros(memory),
+            internal: DMatrix::from_element(1, memory, 0.0),
             evaluator,
             outputs: net.outputs().len(),
         })
@@ -35,18 +31,13 @@ where
 }
 
 impl FeedForwardMatrixFabricator {
-    fn get_arr2(mut dynamic_matrix: Vec<Vec<f64>>) -> Array2<f64> {
-        let dim_x = dynamic_matrix.len();
-        let dim_y = dynamic_matrix[0].len();
+    fn get_matrix(dynamic_matrix: Vec<Vec<f64>>) -> DMatrix<f64> {
+        let columns = dynamic_matrix
+            .into_iter()
+            .map(|vec| DVector::from_vec(vec))
+            .collect::<Vec<_>>();
 
-        let flat_vec = dynamic_matrix.iter_mut().fold(Vec::new(), |mut flat, col| {
-            flat.append(col);
-            flat
-        });
-
-        Array::from_shape_vec((dim_x, dim_y), flat_vec)
-            .unwrap()
-            .reversed_axes()
+        DMatrix::from_columns(&columns)
     }
 }
 
@@ -251,7 +242,7 @@ where
         Ok(super::evaluator::MatrixEvaluator {
             stages: compute_stages
                 .into_iter()
-                .map(FeedForwardMatrixFabricator::get_arr2)
+                .map(FeedForwardMatrixFabricator::get_matrix)
                 .collect(),
             transformations: stage_transformations,
         })
@@ -260,14 +251,14 @@ where
 
 #[cfg(test)]
 mod tests {
+    use nalgebra::dmatrix;
+
     use super::{FeedForwardMatrixFabricator, RecurrentMatrixFabricator};
     use crate::{
         edges,
-        matrix::evaluator::SelfNormalizingMatrixEvaluator,
         network::{net::Net, Evaluator, Fabricator, StatefulEvaluator, StatefulFabricator},
         nodes,
     };
-    use ndarray::array;
 
     // tests construction and evaluation of simplest network
     #[test]
@@ -277,10 +268,10 @@ mod tests {
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
         // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0]);
         // println!("result {:?}", result);
 
-        assert_eq!(result, array![2.5]);
+        assert_eq!(result, dmatrix![2.5]);
     }
 
     // tests input dimension > 1
@@ -299,10 +290,10 @@ mod tests {
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
         // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0, 5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0, 5.0]);
         // println!("result {:?}", result);
 
-        assert_eq!(result, array![5.0]);
+        assert_eq!(result, dmatrix![5.0]);
     }
 
     // test linear chaining of edges
@@ -321,10 +312,10 @@ mod tests {
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
         // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0]);
         // println!("result {:?}", result);
 
-        assert_eq!(result, array![1.25]);
+        assert_eq!(result, dmatrix![1.25]);
     }
 
     // test construction of carry for later needs
@@ -344,10 +335,10 @@ mod tests {
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
         // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0]);
         // println!("result {:?}", result);
 
-        assert_eq!(result, array![3.75]);
+        assert_eq!(result, dmatrix![3.75]);
     }
 
     // test construction of carry for early result with dedup carry
@@ -366,12 +357,10 @@ mod tests {
         );
 
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
-        // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0]);
-        // println!("result {:?}", result);
+        let result = evaluator.evaluate(dmatrix![5.0]);
 
-        assert_eq!(result, array![3.75, 2.5]);
+        assert_eq!(result, dmatrix![3.75, 2.5]);
     }
 
     // test construction of carry for early result flipped order
@@ -391,10 +380,10 @@ mod tests {
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
         // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0]);
         // println!("result {:?}", result);
 
-        assert_eq!(result, array![2.5, 1.25]);
+        assert_eq!(result, dmatrix![2.5, 1.25]);
     }
 
     // test unconnected net
@@ -451,34 +440,10 @@ mod tests {
         let evaluator = FeedForwardMatrixFabricator::fabricate(&some_net).unwrap();
         // println!("stages {:?}", evaluator.stages);
 
-        let result = evaluator.evaluate(array![5.0, 5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0, 5.0]);
         // println!("result {:?}", result);
 
-        assert_eq!(result, array![2.5]);
-    }
-
-    #[test]
-    fn simple_net_evaluator_10() {
-        let some_net = Net::new(
-            2,
-            1,
-            nodes!('l', 'l', 'l'),
-            edges!(
-                0--0.5->2,
-                1--0.5->2
-            ),
-        );
-
-        let mut evaluator: SelfNormalizingMatrixEvaluator =
-            FeedForwardMatrixFabricator::fabricate(&some_net)
-                .unwrap()
-                .into();
-        // println!("stages {:?}", evaluator.stages);
-
-        dbg!(&evaluator);
-
-        assert_eq!(evaluator.evaluate(array![5.0, 5.0]), array![0.0]);
-        assert_eq!(evaluator.evaluate(array![2.5, 7.5]), array![0.0]);
+        assert_eq!(result, dmatrix![2.5]);
     }
 
     #[test]
@@ -498,18 +463,17 @@ mod tests {
             1--1.0->3
         ));
         let mut evaluator = RecurrentMatrixFabricator::fabricate(&some_net).unwrap();
-        // println!("stages {:?}", evaluator);
 
-        let result = evaluator.evaluate(array![5.0, 0.0]);
-        assert_eq!(result, array![5.0, 0.0]);
+        let result = evaluator.evaluate(dmatrix![5.0, 0.0]);
+        assert_eq!(result, dmatrix![5.0, 0.0]);
 
-        let result = evaluator.evaluate(array![5.0, 5.0]);
-        assert_eq!(result, array![10.0, 5.0]);
+        let result = evaluator.evaluate(dmatrix![5.0, 5.0]);
+        assert_eq!(result, dmatrix![10.0, 5.0]);
 
-        let result = evaluator.evaluate(array![0.0, 5.0]);
-        assert_eq!(result, array![5.0, 10.0]);
+        let result = evaluator.evaluate(dmatrix![0.0, 5.0]);
+        assert_eq!(result, dmatrix![5.0, 10.0]);
 
-        let result = evaluator.evaluate(array![0.0, 0.0]);
-        assert_eq!(result, array![0.0, 5.0]);
+        let result = evaluator.evaluate(dmatrix![0.0, 0.0]);
+        assert_eq!(result, dmatrix![0.0, 5.0]);
     }
 }
